"""Connection to HoreKa hardware measurements."""
import platform
from datetime import datetime
from pathlib import Path
from typing import List, Union

import pandas as pd
from influxdb_client import InfluxDBClient

from perun import config
from perun.comm import Comm

query = """from(bucket: "hk-collector")
|> range(start: _start, stop: _stop)
|> filter(fn: (r) => r["hostname"] == _node)
|> filter(fn: (r) => r["_measurement"] == "consumed_watts" or r["_measurement"] == "nv_power_usage")
|> pivot(rowKey: ["_time"], columnKey: ["_measurement"], valueColumn: "_value")"""


def get_horeka_measurements(
    comm: Comm, outdir: Path, expName: str, expId: int, start: datetime, stop: datetime
):
    """
    Read hardware data from an Influx Database.

    Args:
        comm (Comm): MPI Communication Object
        outdir (Path): Result location
        expName (str): Experiment config.get("horeka", "org")
        expId (int): Experiment Id
        start (datetime): Experiment start time
        stop (datetime): Experiment end time
    """
    URL = config.get("horeka", "url")
    TOKEN = config.get("horeka", "token")
    ORG = config.get("horeka", "org")
    nodename = platform.node().replace(".localdomain", "")

    idb = InfluxDBClient(url=URL, token=TOKEN, org=ORG)

    now = datetime.now()
    p = {"_start": start - now, "_stop": stop - now, "_node": nodename}

    outpath = outdir / "horeka" / expName / str(expId)

    if comm.Get_rank() == 0:
        if not outpath.exists():
            outpath.mkdir(parents=True)

    dfList: Union[List[pd.DataFrame], pd.DataFrame] = idb.query_api().query_data_frame(
        query=query, params=p
    )

    if isinstance(dfList, List):
        for index, df in enumerate(dfList):
            df.to_csv(outpath / f"{nodename}_{index}.csv")
    else:
        dfList.to_csv(outpath / f"{nodename}.csv")



"""Defines Intel RAPL related classes."""
import os
import re
from io import IOBase
from pathlib import Path
from typing import Callable, List, Set

import cpuinfo
import numpy as np

from perun import log
from perun.data_model.measurement_type import Magnitude, MetricMetaData, Unit

from ..data_model.sensor import DeviceType, Sensor
from .backend import Backend, backend

RAPL_PATH = "/sys/class/powercap/"

DIR_RGX = r"intel-rapl:(\d)$"
SUBDIR_RGX = r"intel-rapl:\d:\d$"


@backend
class IntelRAPLBackend(Backend):
    """Intel RAPL as a source of cpu and memory devices.

    Uses pyRAPL to gather device information and creates metrics for each available device
    """

    name = "Intel RAPL"
    description = "Reads energy usage from CPUs and DRAM using Intel RAPL"

    def __init__(self) -> None:
        """Init IntelRAPLBackend."""
        super().__init__()
        log.info("Initialized Intel RAPL")

    def setup(self):
        """Check Intel RAPL access."""
        cpuInfo = cpuinfo.get_cpu_info()
        self.metadata = {
            "vendor_id": cpuInfo["vendor_id_raw"],
            "brand": cpuInfo["brand_raw"],
            "arch": cpuInfo["arch_string_raw"],
            "hz_advertised": cpuInfo["hz_advertised_friendly"],
            "hz_actual": cpuInfo["hz_actual_friendly"],
            "l1_data_cache_size": cpuInfo["l1_data_cache_size"],
            "l1_instruction_cache_size": cpuInfo["l1_instruction_cache_size"],
            "l2_cache_size": cpuInfo["l2_cache_size"],
            "l2_cache_line_size": cpuInfo["l2_cache_line_size"],
            "l2_cache_associativity": cpuInfo["l2_cache_associativity"],
            "l3_cache_size": cpuInfo["l3_cache_size"],
            "source": "Intel RAPL",
        }
        log.debug(f"CPU info metadata: {self.metadata}")

        raplPath = Path(RAPL_PATH)

        def getCallback(file: IOBase) -> Callable[[], np.number]:
            def func() -> np.number:
                file.seek(0)
                return np.uint64(file.readline().strip())

            return func

        if not raplPath.exists():
            raise ImportWarning("No powercap interface")

        self._files = []
        packageDevices = []
        foundPsys = False
        for child in raplPath.iterdir():
            log.debug(child)
            match = re.match(DIR_RGX, child.name)
            if match:
                if os.access(child / "energy_uj", os.R_OK):
                    socket = match.groups()[0]
                    with open(child / "name", "r") as file:
                        device_name = file.readline().strip()

                    log.debug(device_name)

                    if "dram" in device_name:
                        devType = DeviceType.RAM
                    elif "package" in device_name:
                        devType = DeviceType.CPU
                    elif "psys" in device_name:
                        devType = DeviceType.CPU
                        foundPsys = True
                    else:
                        devType = DeviceType.OTHER

                    if devType != DeviceType.OTHER:
                        with open(child / "max_energy_range_uj", "r") as file:
                            max_energy = np.uint64(file.readline().strip())
                        dataType = MetricMetaData(
                            Unit.JOULE,
                            Magnitude.MICRO,
                            np.dtype("uint64"),
                            np.uint64(0),
                            max_energy,
                            max_energy,
                        )

                        energy_path = str(child / "energy_uj")
                        energy_file = open(energy_path, "r")
                        self._files.append(energy_file)
                        device = Sensor(
                            f"{devType.value}_{socket}_{device_name}",
                            devType,
                            self.metadata,
                            dataType,
                            getCallback(energy_file),
                        )

                        self.devices[device.id] = device
                        if "package" in device_name:
                            packageDevices.append(device)

                        for grandchild in child.iterdir():
                            match = re.match(SUBDIR_RGX, grandchild.name)
                            if match:
                                with open(grandchild / "name", "r") as file:
                                    device_name = file.readline().strip()
                                if "dram" in device_name:
                                    devType = DeviceType.RAM
                                elif "package" in device_name:
                                    devType = DeviceType.CPU
                                elif "psys" in device_name:
                                    devType = DeviceType.CPU
                                    foundPsys = True
                                else:
                                    devType = DeviceType.OTHER

                                if devType != DeviceType.OTHER:
                                    with open(
                                        child / "max_energy_range_uj", "r"
                                    ) as file:
                                        max_energy = np.uint64(file.readline().strip())

                                    dataType = MetricMetaData(
                                        Unit.JOULE,
                                        Magnitude.MICRO,
                                        np.dtype("uint64"),
                                        np.uint64(0),
                                        max_energy,
                                        max_energy,
                                    )

                                    energy_path = str(grandchild / "energy_uj")
                                    energy_file = open(energy_path, "r")
                                    self._files.append(energy_file)
                                    device = Sensor(
                                        f"{devType.value}_{socket}_{device_name}",
                                        devType,
                                        self.metadata,
                                        dataType,
                                        getCallback(energy_file),
                                    )
                                    log.debug(device)
                                    self.devices[device.id] = device

                                    if "package" in device_name:
                                        packageDevices.append(device)

        if foundPsys:
            for pkg in packageDevices:
                del self.devices[pkg.id]

        log.debug("IntelRapl devices", self.devices)

    def close(self) -> None:
        """Backend shutdown code (does nothing for intel rapl)."""
        log.debug("Closing files")
        for file in self._files:
            file.close()
        return

    def visibleSensors(self) -> Set[str]:
        """
        Return string ids of visible devices.

        Returns:
            Set[str]: Set with device string ids
        """
        log.debug(self.devices)
        return {id for id, device in self.devices.items()}

    def getSensors(self, deviceList: Set[str]) -> List[Sensor]:
        """
        Gather devive objects based on a set of device ids.

        Args:
            deviceList (Set[str]): Set containing devices ids

        Returns:
            List[Device]: Device objects
        """
        return [self.devices[deviceId] for deviceId in deviceList]


IntelRAPLBackend()



"""IO Yaml module."""
from typing import Optional

from perun.data_model.data import DataNode

# import yaml


def exportYaml(
    dataNode: DataNode, depth: Optional[int] = None, include_raw: bool = False
) -> str:
    """Export DataNode to yaml format.

    Args:
        dataNode (DataNode): DataNode
        depth (Optional[int], optional): If specified, export only the first 'depth' levels of the DataNode tree. Defaults to None.
        include_raw (bool, optional): If raw data should be included. Defaults to False.

    Returns:
        str: json string from dataNode
    """
    raise NotImplementedError("Featured is in development.")


def importYaml(yamlStr: str) -> DataNode:
    """Create DataNode from YAML object.

    Args:
        yamlStr (str): YAML string

    Returns:
        DataNode: DataNode
    """
    raise NotImplementedError("Feature not implemented yet.")



"""psutil backend."""
from typing import Callable, List, Set

import numpy as np
import psutil

from perun import log
from perun.backend.backend import Backend, backend
from perun.data_model.measurement_type import Magnitude, MetricMetaData, Unit
from perun.data_model.sensor import DeviceType, Sensor


@backend
class PSUTILBackend(Backend):
    """PSUTIL Backend class."""

    name: str = "PSUTIL"
    description: str = "Obtain hardware data from psutil"

    def __init__(self) -> None:
        """Create psutil backend."""
        super().__init__()
        log.info("Init PSUTIL")

    def setup(self):
        """Configure psutil backend."""
        psutil.disk_io_counters.cache_clear()
        psutil.net_io_counters.cache_clear()
        self.metadata = {"source": f"psutil {psutil.__version__}"}

        for deviceName in self.visibleSensors():
            if deviceName == "RAM_USAGE":
                mem = psutil.virtual_memory()
                self.devices[deviceName] = Sensor(
                    deviceName,
                    DeviceType.RAM,
                    {"total": mem.total, "available": mem.available, **self.metadata},
                    MetricMetaData(
                        Unit.PERCENT,
                        Magnitude.ONE,
                        np.dtype("float32"),
                        np.float32(0),
                        np.float32(1.0),
                        np.float32(-1),
                    ),
                    self._getCallback(deviceName),
                )
            elif deviceName == "CPU_USAGE":
                self.devices[deviceName] = Sensor(
                    deviceName,
                    DeviceType.CPU,
                    {**self.metadata},
                    MetricMetaData(
                        Unit.PERCENT,
                        Magnitude.ONE,
                        np.dtype("float32"),
                        np.float32(0),
                        np.float32(100.0),
                        np.float32(-1),
                    ),
                    self._getCallback(deviceName),
                )
            elif "BYTES" in deviceName:
                self.devices[deviceName] = Sensor(
                    deviceName,
                    DeviceType.DISK if "DISK" in deviceName else DeviceType.NET,
                    {**self.metadata},
                    MetricMetaData(
                        Unit.BYTE,
                        Magnitude.ONE,
                        np.dtype("uint32"),
                        np.uint32(0),
                        np.uint32(np.iinfo("uint32").max),
                        np.uint32(np.iinfo("uint32").max),
                    ),
                    self._getCallback(deviceName),
                )

    def close(self):
        """Close backend."""
        pass

    def visibleSensors(self) -> Set[str]:
        """Return list of visible devices."""
        return {
            "RAM_USAGE",
            "CPU_USAGE",
            "DISK_READ_BYTES",
            "DISK_WRITE_BYTES",
            "NET_WRITE_BYTES",
            "NET_READ_BYTES",
        }

    def _getCallback(self, device: str) -> Callable[[], np.number]:
        """Return measuring function for each device."""
        if device == "RAM_USAGE":

            def func() -> np.number:
                return np.float32(psutil.virtual_memory().percent / 100)

        elif device == "CPU_USAGE":

            def func() -> np.number:
                return np.float32(psutil.cpu_percent())

        elif device == "DISK_READ_BYTES":

            def func() -> np.number:
                return np.uint32(psutil.disk_io_counters(nowrap=True).read_bytes)

        elif device == "DISK_WRITE_BYTES":

            def func() -> np.number:
                return np.uint32(psutil.disk_io_counters(nowrap=True).write_bytes)

        elif device == "NET_WRITE_BYTES":

            def func() -> np.number:
                return np.uint32(psutil.net_io_counters(nowrap=True).bytes_sent)

        elif device == "NET_READ_BYTES":

            def func() -> np.number:
                return np.uint32(psutil.net_io_counters(nowrap=True).bytes_recv)

        else:
            raise ValueError("Invalid device name")

        return func

    def getSensors(self, deviceList: Set[str]) -> List[Sensor]:
        """Return desired device objects."""
        return [self.devices[deviceName] for deviceName in deviceList]


PSUTILBackend()



"""Text report module."""

from prettytable import PrettyTable

from perun import config, log
from perun.data_model.data import DataNode, MetricType, NodeType, Stats
from perun.util import value2str


def textReport(dataNode: DataNode) -> str:
    """
    Create text report from based on root node of the DataNode tree.

    Args:
        dataNode: Root of DataNode tree.

    Returns:
        str: txt report
    """
    if dataNode.type != NodeType.MULTI_RUN and dataNode.type != NodeType.RUN:
        log.error(
            "Text reports are meant for only to get a brief overview of individual runs or of benchmark results over multiple runs."
        )
        raise Exception("Invalid dataNode type.")

    if not dataNode.processed:
        log.error("Data has not been processed, unable to create report.")
        raise Exception("Cannot generate report from unprocessed data node.")

    reportStr = (
        "------------------------------------------\n"
        "PERUN REPORT\n"
        "\n"
        f"App name: {dataNode.metadata['app_name']}\n"
        f"Run ID: {dataNode.id}\n"
    )

    if dataNode.type == NodeType.MULTI_RUN:
        table = PrettyTable(float_format="1.3f")
        table.field_names = ["Name", "mean", "std", "max", "min"]
        table.align = "r"
        table.align["Name"] = "l"  # type: ignore
        for metricType in MetricType:
            if metricType in dataNode.metrics:
                metric = dataNode.metrics[metricType]
                if isinstance(metric, Stats):
                    value, tfactor, mag = value2str(metric.value, metric.metric_md)
                    table.add_row(
                        [
                            f"{metricType.name} [{mag.symbol}{metric.metric_md.unit.value}]",
                            f"{metric.mean / tfactor:.3f}",
                            f"{metric.std / tfactor:.3f}",
                            f"{metric.max / tfactor:.3f}",
                            f"{metric.min / tfactor:.3f}",
                        ]
                    )
        reportStr += "\n"
        reportStr += table.get_string(float_format=".3")

    else:
        for metricType in MetricType:
            if metricType in dataNode.metrics:
                metric = dataNode.metrics[metricType]
                value, _, mag = value2str(metric.value, metric.metric_md)
                reportStr += f"{metricType.name}: {value} {mag.symbol}{metric.metric_md.unit.value}\n"

        if MetricType.ENERGY in dataNode.metrics:
            e_metric = dataNode.metrics[MetricType.ENERGY]
            e_pue = e_metric.value * config.getfloat("post-processing", "pue")
            e_kWh = e_pue / (3600 * 1e3)
            kgCO2 = e_kWh * config.getfloat("post-processing", "emissions_factor") / 1e3
            euro = e_kWh * config.getfloat("post-processing", "price_factor") / 1e2

            reportStr += f"\nThe application used a total of {e_kWh:.3f} kWh, released a total of {kgCO2:.3f} kgCO2eq into the atmosphere, and you paid {euro:.2f} € for it.\n"

    return reportStr



"""Util module."""
import os
from datetime import datetime, timedelta
from pathlib import Path
from typing import Callable, Tuple, Union

import numpy as np

from perun import config
from perun.data_model.measurement_type import Magnitude, MetricMetaData, Unit


def getRunName(app: Union[Path, Callable]) -> str:
    """Return application name based on available info.

    Args:
        app (Union[Path, Callable]): The path or function that is being monitored.

    Returns:
        str: Application name.
    """
    if config.get("output", "app_name"):
        return config.get("output", "app_name")
    elif "SBATCH_JOB_NAME" in os.environ:
        return os.environ["SBATCH_JOB_NAME"]
    elif isinstance(app, Path):
        return app.stem
    else:
        return app.__name__


def getRunId(startime: datetime) -> str:
    """Get run id from available info.

    Args:
        startime (datetime): Time when application was started.

    Returns:
        str: String id.
    """
    if config.get("output", "run_id"):
        return config.get("output", "run_id")
    elif "SLURM_JOB_ID" in os.environ:
        return os.environ["SLURM_JOB_ID"]
    else:
        return startime.isoformat()


def value2str(
    value: np.number, metric_md: MetricMetaData
) -> Tuple[str, float, Magnitude]:
    """Return a printable representation of the value based on its metric metadata. A printable value should not have more than 3 digits after before the decimal comma/dot.

    Args:
        value (np.number): Value to format.
        metric_md (MetricMetaData): Value metadata.

    Returns:
        Tuple[str, float, Magnitude]: Formated value, transformation factor used, and the new magnitude prefix.
    """
    if metric_md.unit == Unit.WATT or metric_md.unit == Unit.JOULE:
        transformFactor = 1
        for mag in reversed(Magnitude):
            if value > mag.value:
                transformFactor = mag.value
                break

        newValue = value / transformFactor
        newMag = Magnitude(metric_md.mag.value * transformFactor)
        return f"{newValue:.3f}", transformFactor, newMag

    elif metric_md.unit == Unit.PERCENT:
        return f"{value:.3f}", 1.0, metric_md.mag
    elif metric_md.unit == Unit.SECOND:
        return str(timedelta(seconds=float(value))), 1.0, Magnitude.ONE
    elif metric_md.unit == Unit.BYTE:
        transformFactor = 1
        newMag = Magnitude.ONE
        for magFactor, m in zip(
            [1024**3, 1024**2, 1024**1],
            [Magnitude.GIGA, Magnitude.MEGA, Magnitude.KILO],
        ):
            if value > magFactor:
                transformFactor = magFactor
                newMag = m
                break

        newValue = value / transformFactor
        return f"{newValue:.3f}", transformFactor, newMag
    else:
        return f"{value:.3f}", 1.0, metric_md.mag



"""Bench io module."""
import json

from perun.data_model.data import DataNode, MetricType, Stats
from perun.util import value2str

lessIsBetterMetrics = [MetricType.RUNTIME, MetricType.ENERGY]


def exportBench(dataNode: DataNode) -> str:
    """Export data node to json format based on the github continuous benchmark action.

    https://github.com/benchmark-action/github-action-benchmark

    Args:
        dataNode (DataNode): Perun results

    Returns:
        str: Output string
    """
    metricDict = []
    for metricType, metric in dataNode.metrics.items():
        if metricType in lessIsBetterMetrics:
            metric_md = metric.metric_md
            value, tfactor, mag = value2str(metric.value, metric_md)

            if isinstance(metric, Stats):
                metricDict.append(
                    {
                        "name": f"{dataNode.metadata['app_name']} - {metricType.name}",
                        "unit": f"{mag.symbol}{metric_md.unit.symbol}",
                        "value": metric.mean / tfactor,
                        "range": metric.std / tfactor,
                    }
                )
            else:
                metricDict.append(
                    {
                        "name": f"{dataNode.metadata['app_name']} - {metricType.name}",
                        "unit": f"{mag.symbol}{metric_md.unit.symbol}",
                        "value": metric.value / tfactor,
                    }
                )

    return json.dumps(metricDict, indent=4)



"""Backend module."""
import functools
from abc import ABC, abstractmethod
from typing import Dict, List, Set

from perun import log

from ..data_model.sensor import Sensor


class Backend(ABC):
    """Abstract backend class."""

    name: str = "Abstract backend class"
    description: str = "Abstract backend class description"

    def __init__(self) -> None:
        """Import and setup backend."""
        super().__init__()
        self.devices: Dict = {}
        self.setup()

    @abstractmethod
    def visibleSensors(self) -> Set[str]:
        """Get a string id of devices visible by the backend."""
        pass

    @abstractmethod
    def getSensors(self, deviceList: Set[str]) -> List[Sensor]:
        """
        Return device objects based on the provided list of device ids.

        Args:
            deviceList (Set[str]): List with wanted device ids

        Returns:
            List[Device]: List of device objects
        """
        pass

    @abstractmethod
    def close(self):
        """Clean up and close backend related activities."""
        pass

    @abstractmethod
    def setup(self):
        """Perform backend setup."""
        pass


backends: List[Backend] = []


def backend(cls):
    """Backend class decorator.

    Marks a class a singleton, and if setup succeeds,
    gets added to the backends list.
    """

    @functools.wraps(cls)
    def backend_wrapper(*args, **kwargs):
        if not backend_wrapper.instance:
            try:
                backend_wrapper.instance = cls(*args, **kwargs)
                backends.append(backend_wrapper.instance)
            except ImportError as ie:
                log.warning(f"Missing dependencies for backend {cls.__name__}")
                log.warning(ie)
            except Exception as e:
                log.warning(f"Unknown error loading dependecy {cls.__name__}")
                log.warning(e)

    backend_wrapper.instance = None
    return backend_wrapper



"""Nvidia Mangement Library Source definition."""
from typing import Callable, Set

import numpy as np
import pynvml
from pynvml import NVMLError

from perun import log
from perun.data_model.measurement_type import Magnitude, MetricMetaData, Unit

from ..data_model.sensor import DeviceType, Sensor
from .backend import Backend, backend


@backend
class NVMLBackend(Backend):
    """NVMLSource class.

    Setups connection to NVML and creates relevant devices
    """

    name = "NVIDIA ML"
    description: str = "Access GPU information from NVML python bindings"

    def __init__(self) -> None:
        """Init NVIDIA ML Backend."""
        super().__init__()
        log.info("Initialized NVML Backend")

    def setup(self):
        """Init pynvml and gather number of devices."""
        pynvml.nvmlInit()
        deviceCount = pynvml.nvmlDeviceGetCount()
        self.metadata = {
            "cuda_version": pynvml.nvmlSystemGetCudaDriverVersion(),
            "driver_version": pynvml.nvmlSystemGetDriverVersion(),
            "source": "Nvidia Managment Library",
        }

        log.debug(f"NVML Device count: {deviceCount}")

    def close(self):
        """Backend shutdown code."""
        pynvml.nvmlShutdown()

    def visibleSensors(self) -> Set[str]:
        """
        Return string ids of visible devices.

        Returns:
            Set[str]: Set with device string ids
        """
        devices = set()
        for i in range(pynvml.nvmlDeviceGetCount()):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            devices.add(pynvml.nvmlDeviceGetUUID(handle))
        return devices

    def getSensors(self, deviceList: Set[str]):
        """
        Gather device objects based on a set of device ids.

        Args:
            deviceList (Set[str]): Set containing devices ids

        Returns:
            List[Device]: Device objects
        """

        def getCallback(handle) -> Callable[[], np.number]:
            def func() -> np.number:
                return np.uint32(pynvml.nvmlDeviceGetPowerUsage(handle))

            return func

        devices = []
        for deviceId in deviceList:
            try:
                handle = pynvml.nvmlDeviceGetHandleByUUID(deviceId.encode())
                index = pynvml.nvmlDeviceGetIndex(handle)

                name = f"CUDA:{index}"
                device_type = DeviceType.GPU
                device_metadata = {
                    "uuid": deviceId,
                    "name": pynvml.nvmlDeviceGetName(handle),
                    **self.metadata,
                }
                max_power = np.uint32(pynvml.nvmlDeviceGetPowerManagementLimit(handle))

                data_type = MetricMetaData(
                    Unit.WATT,
                    Magnitude.MILI,
                    np.dtype("uint32"),
                    np.uint32(0),
                    np.uint32(max_power),
                    np.uint32(0),
                )
                devices.append(
                    Sensor(
                        name,
                        device_type,
                        device_metadata,
                        data_type,
                        getCallback(handle),
                    )
                )
            except NVMLError as e:
                log.debug(f"Could not find device {deviceId}")
                log.debug(e)

        return devices


NVMLBackend()



"""perun module."""
# flake8: noqa
__version__ = "0.1.1"
from perun.configuration import config
from perun.logging import init_logging

log = init_logging(config.get("debug", "log_lvl"))

import os

os.environ["IBV_FORK_SAFE"] = "1"
os.environ["RDMAV_FORK_SAFE"] = "1"

from perun.comm import Comm

COMM_WORLD = Comm()



"""Decorator module."""

import functools

from perun import config, log
from perun.configuration import read_custom_config, save_to_config
from perun.perun import monitor_application


def monitor(
    configuration: str = "./.perun.ini",
    **conf_kwargs,
):
    """Decorate function to monitor its energy usage."""

    def inner_function(func):
        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):
            # Get custom config and kwargs

            read_custom_config(None, None, configuration)
            for key, value in conf_kwargs.items():
                save_to_config(key, value)

            log.setLevel(config.get("debug", "log_lvl"))
            func_result = monitor_application(func, args, kwargs)

            return func_result

        return func_wrapper

    return inner_function



"""Comm module."""
from typing import Any, List, Optional

from perun import log


class Comm:
    """Wrapper around MPI COMM_WORLD. Does nothing if MPI is not initialized."""

    def __init__(self):
        """Initialize MPI Communicator if availablel."""
        self._enabled = False
        self._rank = 0
        self._size = 1
        try:
            from mpi4py import MPI

            if MPI.COMM_WORLD.Get_size() >= 1:
                self._comm = MPI.COMM_WORLD
                self._enabled = True

        except ImportError as e:
            log.warning("Missing mpi4py, multi-node monitoring disabled")
            log.warning(e)

    def Get_rank(self) -> int:
        """Return MPI rank.

        Returns:
            int: MPI Rank
        """
        return self._comm.Get_rank() if self._enabled else self._rank

    def Get_size(self) -> int:
        """Return MPI world size.

        Returns:
            int: MPI world size
        """
        return self._comm.Get_size() if self._enabled else self._size

    def gather(self, obj: Any, root: int = 0) -> Optional[List[Any]]:
        """MPI gather operation at selected rank.

        Args:
            obj (Any): Object to be gathererd.
            root (int, optional): Rank to gather information at. Defaults to 0.

        Returns:
            Optional[List[Any]]: List with objects from all the ranks.
        """
        return self._comm.gather(obj, root=root) if self._enabled else [obj]

    def allgather(self, obj: Any) -> List[Any]:
        """MPI allgather operation."""
        if self._enabled:
            return self._comm.allgather(obj)
        else:
            return [obj]

    def bcast(self, obj: Any, root: int = 0) -> Any:
        """MPI broadcast operation."""
        if self._enabled:
            return self._comm.bcast(obj, root)
        else:
            return obj

    def barrier(self):
        """MPI barrier operation."""
        if self._enabled:
            self._comm.barrier()



"""IO Module."""

import enum
from pathlib import Path
from typing import Optional, Union

from perun import log
from perun.data_model.data import DataNode, NodeType
from perun.io.bench import exportBench
from perun.io.hdf5 import exportHDF5, importHDF5
from perun.io.json import exportJson, importJson
from perun.io.pandas import exportCSV
from perun.io.pickle import exportPickle, importPickle
from perun.io.text_report import textReport

_suffixes = {
    "text": "txt",
    "yaml": "yaml",
    "json": "json",
    "hdf5": "hdf5",
    "pickle": "pkl",
    "csv": "csv",
    "bench": "json",
}


class IOFormat(enum.Enum):
    """Available IO file formats."""

    TEXT = "text"
    JSON = "json"
    HDF5 = "hdf5"
    PICKLE = "pickle"
    CSV = "csv"
    BENCH = "bench"
    # YAML = "yaml"

    @property
    def suffix(self):
        """Return file suffix from format."""
        return _suffixes[self.value]

    @classmethod
    def fromSuffix(cls, suffix: str):
        """Return format from suffix."""
        for key, value in _suffixes.items():
            if value in suffix:
                return cls(key)
        raise ValueError("Invalid file format.")


def exportTo(
    data_out: Path,
    dataNode: DataNode,
    format: IOFormat,
    rawData: bool = False,
    depth: Optional[int] = None,
):
    """Export DataNode structure to the selected format.

    Args:
        dataNode (DataNode): DataNode tree with processed metrics.
        format (IOFormat, optional): Output format. Defaults to IOFormat.TEXT.
        rawData (bool, optional): If raw data should be included. Limits available formats. Defaults to False.
    """
    if not dataNode.processed:
        log.warning("Data has not been processed before import. Proceed with caution.")
        raise Exception("DataNode needs to be processed before it can be exported.")

    if not data_out.exists():
        log.info(f"{data_out} does not exists. So lets make it.")
        data_out.mkdir()

    if format == IOFormat.BENCH and dataNode.type != NodeType.MULTI_RUN:
        log.warning(
            "BENCH format can only be used with 'bench' mode enabled. Using pickle instead."
        )
        format = IOFormat.PICKLE

    filename = f"{dataNode.metadata['app_name']}_{dataNode.id}"

    reportStr: Union[str, bytes]
    if format == IOFormat.JSON:
        filename += ".json"
        fileType = "w"
        reportStr = exportJson(dataNode, depth, rawData)
        with open(data_out / filename, fileType) as file:
            file.write(reportStr)
    elif format == IOFormat.HDF5:
        filename += ".hdf5"
        exportHDF5(data_out / filename, dataNode)
    elif format == IOFormat.PICKLE:
        filename += ".pkl"
        fileType = "wb"
        reportStr = exportPickle(dataNode)
        with open(data_out / filename, fileType) as file:
            file.write(reportStr)
    elif format == IOFormat.CSV:
        filename += ".cvs"
        exportCSV(data_out / filename, dataNode)
    elif format == IOFormat.BENCH:
        filename += ".json"
        fileType = "w"
        reportStr = exportBench(dataNode)
        with open(data_out / filename, fileType) as file:
            file.write(reportStr)
    # elif format == IOFormat.YAML:
    #     filename += ".yaml"
    #     reportStr = exportYaml(dataNode, depth, rawData)
    else:
        filename += ".txt"
        fileType = "w"
        reportStr = textReport(dataNode)
        with open(data_out / filename, fileType) as file:
            file.write(reportStr)


def importFrom(filePath: Path, format: IOFormat) -> DataNode:
    """Import DataNode structure from path. If no format is given, it is inferred form path sufix.

    Args:
        filePath (Path): Path to file
        format (Optional[IOFormat], optional): File format. Defaults to None.
    """
    if format == IOFormat.JSON:
        with open(filePath, "r") as file:
            dataNode = importJson(file.read())
    # elif format == IOFormat.YAML:
    #     filename += ".yaml"
    #     reportStr = exportYaml(dataNode, depth, rawData)
    elif format == IOFormat.HDF5:
        dataNode = importHDF5(filePath)
    elif format == IOFormat.PICKLE:
        with open(filePath, "rb") as file:
            dataNode = importPickle(file.read())
    else:
        raise ValueError("File format is not supported.")

    return dataNode



"""IO Pickle module."""
import pickle

from perun.data_model.data import DataNode


def exportPickle(dataNode: DataNode) -> bytes:
    """Pickle data node.

    Args:
        dataNode (DataNode): DataNode to be pickled.

    Returns:
        bytes: Pickled DataNode
    """
    return pickle.dumps(dataNode)


def importPickle(pickleData: bytes) -> DataNode:
    """Unpickle DataNode.

    Args:
        pickleData (bytes): Pickled DataNode

    Returns:
        DataNode: Unpickled DataNode
    """
    return pickle.loads(pickleData)



"""IO Json module."""
import json
from typing import Optional

import numpy as np

from perun.data_model.data import DataNode


class NumpyEncoder(json.JSONEncoder):
    """Json Numpy object encoder."""

    def default(self, obj):
        """Transform numpy objects.

        Args:
            obj (_type_): Numpy object
        """
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.dtype):
            return str(obj)
        else:
            return super(NumpyEncoder, self).default(obj)


def exportJson(
    dataNode: DataNode, depth: Optional[int] = None, include_raw: bool = False
) -> str:
    """Export DataNode to json.

    Args:
        dataNode (DataNode): DataNode
        depth (Optional[int], optional): If specified, export only the first 'depth' levels of the DataNode tree. Defaults to None.
        include_raw (bool, optional): If raw data should be included. Defaults to False.

    Returns:
        str: json string from dataNode
    """
    dataDict = dataNode.toDict(depth, include_raw)
    return json.dumps(dataDict, indent=4, cls=NumpyEncoder)


def importJson(jsonString: str) -> DataNode:
    """Create DataNode from JSON string."""
    return DataNode.fromDict(json.loads(jsonString))



"""Command line api definition.

Uses click https://click.palletsprojects.com/en/8.1.x/ to manage complex cmdline configurations.
"""
from pathlib import Path
from pprint import pprint

import click

import perun
from perun import log
from perun.configuration import config, read_custom_config, save_to_config_callback
from perun.io.io import IOFormat, exportTo, importFrom


@click.group()
@click.version_option(version=perun.__version__)
@click.option(
    "-c",
    "--configuration",
    default="./.perun.ini",
    help="Path to configuration file",
    type=click.Path(exists=False, dir_okay=False, readable=True),
    is_eager=True,
    callback=read_custom_config,
    expose_value=False,
)
# Output option
@click.option(
    "-n",
    "--app_name",
    help="Name of the monitored application. The name is used to distinguish between multiple applications in the same directory. If left empty, the filename will be  used.",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "-i",
    "--run_id",
    help="Unique id of the latest run of the application. If left empty, perun will use the SLURM job id, or the current date.",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "--format",
    type=click.Choice([format.value for format in IOFormat]),
    help="Report format.",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "--data_out",
    type=click.Path(exists=False, dir_okay=True, file_okay=False),
    help="Where to save the output files, defaults to the current working directory.",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "--raw",
    default=False,
    help="Use the flag '--raw' if you need access to all the raw data collected by perun. The output will be saved on an hdf5 file on the perun data output location.",
    is_flag=True,
    callback=save_to_config_callback,
    expose_value=False,
)
# Sampling Options
@click.option(
    "--sampling_rate",
    type=float,
    help="Sampling rate in seconds.",
    callback=save_to_config_callback,
    expose_value=False,
)
# Post processing options
@click.option(
    "--pue",
    type=float,
    help="Data center Power Usage Efficiency.",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "--emissions_factor",
    type=float,
    help="Emissions factor at compute resource location.",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "--price_factor",
    type=float,
    help="Electricity price factor at compute resource location.",
    callback=save_to_config_callback,
    expose_value=False,
)
# Benchmarking
@click.option(
    "--bench",
    "bench_enable",
    is_flag=True,
    help="Activate benchmarking mode.",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "--bench_rounds",
    type=int,
    help="Number of rounds per function/app.",
    callback=save_to_config_callback,
    expose_value=False,
)
@click.option(
    "--bench_warmup_rounds",
    type=int,
    help="Number of warmup rounds per function/app.",
    callback=save_to_config_callback,
    expose_value=False,
)
# @click.option(
#     "--bench_metrics",
#     multiple=True,
#     help="Metrics to output. Only relevant with bench_minimal_format enabled",
#     callback=save_to_config_callback,
#     expose_value=False,
# )
# Debug Options
@click.option(
    "-l",
    "--log_lvl",
    type=click.Choice(["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"]),
    help="Loggging level",
    callback=save_to_config_callback,
    expose_value=False,
)
def cli():
    """Perun: Energy measuring and reporting tool."""
    log.setLevel(config.get("debug", "log_lvl"))


@cli.command()
@click.option(
    "--default",
    is_flag=True,
    show_default=True,
    default=False,
    help="Print default configuration",
)
def showconf(default: bool):
    """Print current perun configuration in INI format."""
    import sys

    from perun.configuration import _default_config

    if default:
        config.read_dict(_default_config)
        config.write(sys.stdout)
    else:
        config.write(sys.stdout)


@cli.command()
def sensors():
    """Print sensors assigned to each rank by perun."""
    from perun import COMM_WORLD
    from perun.backend import backends
    from perun.coordination import getGlobalSensorRankConfiguration

    globalHostRank, globalSensorConfig = getGlobalSensorRankConfiguration(
        COMM_WORLD, backends
    )
    if COMM_WORLD.Get_rank() == 0:
        for rank, bes in enumerate(globalSensorConfig):
            click.echo(f"Rank: {rank}")
            click.echo(pprint(bes))

        click.echo("Hostnames: ")
        click.echo(pprint(globalHostRank))

    for b in backends:
        b.close()


@cli.command()
@click.argument("input_file", type=click.Path(exists=True))
@click.argument("output_path", type=click.Path(exists=True))
@click.argument(
    "output_format",
    type=click.Choice([format.value for format in IOFormat]),
)
def export(input_file: str, output_path: str, output_format: str):
    """Export existing perun output file to another format."""
    in_file = Path(input_file)
    if not in_file.exists():
        click.echo("File does not exist.", err=True)
        return

    out_path = Path(output_path)
    if not out_path.parent.exists():
        click.echo("Output path does not exist", err=True)
        return

    inputFormat = IOFormat.fromSuffix(in_file.suffix)
    out_format = IOFormat(output_format)
    dataNode = importFrom(in_file, inputFormat)
    exportTo(out_path, dataNode, out_format, rawData=True)


@cli.command(context_settings={"ignore_unknown_options": True})
@click.argument("script", type=click.Path(exists=True))
@click.argument("script_args", nargs=-1)
def monitor(
    script: str,
    script_args: tuple,
):
    """
    Gather power consumption from hardware devices while SCRIPT [SCRIPT_ARGS] is running.

    SCRIPT is a path to the python script to monitor, run with arguments SCRIPT_ARGS.
    """
    # Setup script arguments
    import sys

    from perun.perun import monitor_application

    filePath: Path = Path(script)
    log.debug(f"Script path: {filePath}")
    argIndex = sys.argv.index(script)
    sys.argv = sys.argv[argIndex:]
    log.debug(f"Script args: { sys.argv }")

    sys.path.insert(0, str(filePath.parent.absolute()))

    monitor_application(filePath)


def main():
    """Cli entrypoint."""
    cli(auto_envvar_prefix="PERUN")



"""Pandas IO module."""
from pathlib import Path
from typing import Any, List, Tuple

import pandas as pd

from perun.data_model.data import DataNode, NodeType


def exportCSV(outputPath: Path, dataNode: DataNode):
    """Export all raw data collected to a csv file using pandas.

    Args:
        outputPath (Path): Output path
        dataNode (DataNode): Perun data node with raw data.
    """
    df = _dataNode2Pandas(dataNode)
    df.to_csv(outputPath)


def _dataNode2Pandas(dataNode: DataNode) -> pd.DataFrame:
    """Create a pandas dataframe for Data Node raw data.

    Args:
        dataNode (DataNode): Perun Data Node

    Returns:
        pd.DataFrame: DataFrame
    """
    columns = [
        "app_name",
        "run_id",
        "hostname",
        "device_group",
        "sensor",
        "unit",
        "magnitude",
        "timestep",
        "value",
    ]

    rows = []
    exp_name = dataNode.metadata["app_name"]
    if dataNode.type == NodeType.MULTI_RUN:
        for id, node in dataNode.nodes.items():
            rows.extend(_rowsFromRunNode(node, exp_name, id))

    else:
        id = dataNode.id
        rows = _rowsFromRunNode(dataNode, exp_name, id)

    return pd.DataFrame(rows, columns=columns)


def _rowsFromRunNode(
    runNode: DataNode, app_name: str, run_id: str
) -> List[Tuple[Any, ...]]:
    """Read raw data from data nodes and arrange it as rows.

    Args:
        runNode (DataNode): Data node.
        app_name (str): Application name.
        run_id (str): Run id.

    Returns:
        List[Tuple[Any]]: Rows with raw data.
    """
    rows: List[Tuple[Any, ...]] = []
    for hostId, hostNode in runNode.nodes.items():
        for deviceGroupId, deviceGroupNode in hostNode.nodes.items():
            for sensorId, sensorNode in deviceGroupNode.nodes.items():
                if sensorNode.raw_data is not None:
                    rawData = sensorNode.raw_data
                    unit = rawData.v_md.unit.value
                    mag = rawData.v_md.mag.value
                    for i in range(len(rawData.timesteps)):
                        rows.append(
                            (
                                app_name,
                                run_id,
                                hostId,
                                deviceGroupId,
                                sensorId,
                                unit,
                                mag,
                                rawData.timesteps[i],
                                rawData.values[i],
                            )
                        )

    return rows



"""Coordination module."""
import platform
from typing import Dict, List, Optional, Set, Tuple

from perun import Comm, log
from perun.backend import Backend

_cached_sensor_config: Optional[Tuple[Dict, List]] = None


def getHostRankDict(comm: Comm) -> Dict[str, List[int]]:
    """Return a dictionary with all the host names with each MPI rank in them.

    Args:
        comm (Comm): MPI Communicator

    Returns:
        Dict[str, List[int]]: Dictionary with key hostnames and mpi ranks as values.
    """
    hostname = platform.node()
    rank = comm.Get_rank()

    gHostRank: List[Tuple[str, int]] = comm.allgather((hostname, rank))
    hostRankDict: Dict[str, List[int]] = {}
    for h, r in gHostRank:
        if h in hostRankDict:
            hostRankDict[h].append(r)
        else:
            hostRankDict[h] = [r]

    return hostRankDict


def getGlobalSensorRankConfiguration(
    comm: Comm, backends: List[Backend]
) -> Tuple[Dict[str, List[int]], List[Dict[str, Set[str]]]]:
    """Gather available sensor information from every MPI rank and assign/unassign sensors to each rank to avoid over sampling.

    Args:
        comm (Comm): MPI Communicator
        backends (List[Backend]): List of available backends in the current rank.

    Returns:
        Tuple[Dict[str, List[int]], List[Dict[str, Set[str]]]]: Global rank and sensor configuration objects.
    """
    visibleSensorsByBackend: Dict[str, Set[str]] = {
        backend.name: backend.visibleSensors() for backend in backends
    }
    log.debug(f"Rank {comm.Get_rank()} : Visible devices {visibleSensorsByBackend}")
    globalVisibleSensorsByBackend = comm.allgather(visibleSensorsByBackend)
    globalHostRanks = getHostRankDict(comm)

    globalSensorConfig = assignSensors(globalVisibleSensorsByBackend, globalHostRanks)
    return (
        globalHostRanks,
        globalSensorConfig,
    )


def getLocalSensorRankConfiguration(
    comm: Comm, backends: List[Backend]
) -> Tuple[List[int], Dict[str, Set[str]]]:
    """Obtain local sensor configuration.

    Args:
        comm (Comm): MPI Communicator
        backends (List[Backend]): List of availbale backends in the current ranks.

    Returns:
        Tuple[List[int], Dict[str, Set[str]]]: Local rank and sensor configuration
    """
    global _cached_sensor_config
    if _cached_sensor_config is None:
        globalHostRanks, globalSensorConfig = getGlobalSensorRankConfiguration(
            comm, backends
        )
        _cached_sensor_config = (globalHostRanks, globalSensorConfig)
    else:
        globalHostRanks, globalSensorConfig = _cached_sensor_config

    return globalHostRanks[platform.node()], globalSensorConfig[comm.Get_rank()]


def assignSensors(
    hostBackends: List[Dict[str, Set[str]]], hostNames: Dict[str, List[int]]
) -> List[Dict[str, Set[str]]]:
    """Assign found devices to the lowest rank in each host.

    Args:
        hostSensors (List[Set[str]]): List with lenght of the mpi world size, with each index containing the devices of each rank.
        hostNames (List[str]): Hostname of the mpi rank at the index.

    Returns:
        List[Set[str]]: New list with the devices assiged to each rank.
    """
    for host, ranks in hostNames.items():
        firstRank = ranks[0]
        firstRankDict = hostBackends[firstRank]
        for rank in ranks[1:]:
            firstRankDict = _mergeDicts(firstRankDict, hostBackends[rank])
            hostBackends[rank] = {}

        hostBackends[firstRank] = firstRankDict
    return hostBackends


def _mergeDicts(
    dict1: Dict[str, Set[str]], dict2: Dict[str, Set[str]]
) -> Dict[str, Set[str]]:
    for key, value in dict2.items():
        if key in dict1:
            dict1[key] |= value
        else:
            dict1[key] = value

    return dict1



"""Logging utilities."""

import logging
import logging.config


def init_logging(level: str):
    """Initialize default stdout logger.

    Args:
        level (str, optional): Desired log level. Defaults to "DEBUG".
    """
    logConfig = {
        "version": 1,
        "root": {"handlers": ["console"], "level": level},
        "handlers": {
            "console": {
                "formatter": "std_out",
                "class": "logging.StreamHandler",
                "level": "DEBUG",
            }
        },
        "formatters": {
            "std_out": {
                "format": "%(asctime)s : %(levelname)s : %(module)s : %(funcName)s -- %(message)s",
                "datefmt": "%d-%m-%Y %I:%M:%S",
            }
        },
    }
    logging.config.dictConfig(logConfig)
    return logging.getLogger("perun")



"""Unit module."""
import dataclasses
import enum
from typing import Dict

import numpy as np
from typing_extensions import Self


class Unit(str, enum.Enum):
    """Unit enum."""

    JOULE = "J"
    WATT = "W"
    BYTE = "B"
    SECOND = "s"
    PERCENT = "%"

    @property
    def symbol(self) -> str:
        """Symbol associated with Unit.

        Returns:
            str: String with unit symbol.
        """
        return self.value

    def __str__(self) -> str:
        """Convert object to string."""
        return self.symbol


class Magnitude(float, enum.Enum):
    """Magnitude prefix enum."""

    PICO = 1e-12
    NANO = 1e-9
    MICRO = 1e-6
    MILI = 1e-3
    ONE = 1
    KILO = 1e3
    MEGA = 1e6
    GIGA = 1e9
    TERA = 1e12

    @property
    def symbol(self) -> str:
        """Symbol associated with magnitude prefix.

        Returns:
            str: String with magnitude symbol.
        """
        _symbols: Dict = {
            "PICO": "p",
            "NANO": "n",
            "MICRO": "µ",
            "MILI": "m",
            "ONE": "",
            "KILO": "k",
            "MEGA": "M",
            "GIGA": "G",
            "TERA": "T",
        }
        return _symbols[self.name]

    def __str__(self) -> str:
        """Convert object to string."""
        return self.symbol


@dataclasses.dataclass
class MetricMetaData:
    """Collects a metric metadata."""

    unit: Unit
    mag: Magnitude
    dtype: np.dtype
    min: np.number
    max: np.number
    fill: np.number

    @classmethod
    def fromDict(cls, mdDict: Dict) -> Self:
        """Create MetricMetadata from a dictionary."""
        dtype = np.dtype(mdDict["dtype"])
        return cls(
            Unit(mdDict["unit"]),
            Magnitude(mdDict["mag"]),
            dtype,
            dtype.type(mdDict["min"]),
            dtype.type(mdDict["max"], dtype=dtype),
            dtype.type(mdDict["fill"], dtype=dtype),
        )



"""Processing Module."""
from typing import Dict, List

import numpy as np

from perun.data_model.data import (
    AggregateType,
    DataNode,
    Metric,
    MetricType,
    NodeType,
    Stats,
)
from perun.data_model.measurement_type import Magnitude, MetricMetaData, Unit
from perun.data_model.sensor import DeviceType


def processSensorData(sensorData: DataNode) -> DataNode:
    """Calculate metrics based on the data found on sensor nodes.

    Args:
        sensorData (DataNode): DataNode with raw data (SENSOR)
    """
    if sensorData.type == NodeType.SENSOR and sensorData.raw_data:
        rawData = sensorData.raw_data

        runtime = rawData.timesteps[-1]
        sensorData.metrics[MetricType.RUNTIME] = Metric(
            MetricType.RUNTIME, runtime, rawData.t_md, AggregateType.MAX
        )

        if rawData.v_md.unit == Unit.JOULE:
            t_s = rawData.timesteps.astype("float32")
            t_s *= rawData.t_md.mag.value / Magnitude.ONE.value

            e_J = rawData.values
            maxValue = rawData.v_md.max
            dtype = rawData.v_md.dtype.name
            d_energy = e_J[1:] - e_J[:-1]
            if "uint" in dtype:
                idx = d_energy >= maxValue
                max_dtype = np.iinfo(dtype).max
                d_energy[idx] = maxValue + d_energy[idx] - max_dtype
            else:
                idx = d_energy <= 0
                d_energy[idx] = d_energy[idx] + maxValue
            total_energy = d_energy.sum()

            magFactor = rawData.v_md.mag.value / Magnitude.ONE.value
            energy_J = np.float32(total_energy) * magFactor

            sensorData.metrics[MetricType.ENERGY] = Metric(
                MetricType.ENERGY,
                energy_J,
                MetricMetaData(
                    Unit.JOULE,
                    Magnitude.ONE,
                    np.dtype("float32"),
                    np.float32(0),
                    np.finfo("float32").max,
                    np.float32(-1),
                ),
                AggregateType.SUM,
            )
            sensorData.metrics[MetricType.POWER] = Metric(
                MetricType.POWER,
                energy_J / runtime,
                MetricMetaData(
                    Unit.WATT,
                    Magnitude.ONE,
                    np.dtype("float32"),
                    np.float32(0),
                    np.finfo("float32").max,
                    np.float32(-1),
                ),
                AggregateType.SUM,
            )
        elif rawData.v_md.unit == Unit.WATT:
            t_s = rawData.timesteps.astype("float32")
            t_s *= rawData.t_md.mag.value / Magnitude.ONE.value

            magFactor = rawData.v_md.mag.value / Magnitude.ONE.value
            power_W = rawData.values.astype("float32") * magFactor
            energy_J = np.trapz(power_W, t_s)
            sensorData.metrics[MetricType.ENERGY] = Metric(
                MetricType.ENERGY,
                energy_J,
                MetricMetaData(
                    Unit.JOULE,
                    Magnitude.ONE,
                    np.dtype("float32"),
                    np.float32(0),
                    np.finfo("float32").max,
                    np.float32(-1),
                ),
                AggregateType.SUM,
            )
            sensorData.metrics[MetricType.POWER] = Metric(
                MetricType.POWER,
                np.mean(power_W),
                MetricMetaData(
                    Unit.WATT,
                    Magnitude.ONE,
                    np.dtype("float32"),
                    np.float32(0),
                    np.finfo("float32").max,
                    np.float32(-1),
                ),
                AggregateType.SUM,
            )
        elif rawData.v_md.unit == Unit.PERCENT:
            if sensorData.deviceType == DeviceType.CPU:
                metricType = MetricType.CPU_UTIL
            elif sensorData.deviceType == DeviceType.GPU:
                metricType = MetricType.GPU_UTIL
            else:
                metricType = MetricType.MEM_UTIL

            sensorData.metrics[metricType] = Metric(
                metricType,
                np.mean(rawData.values),
                rawData.v_md,
                AggregateType.MEAN,
            )
        elif rawData.v_md.unit == Unit.BYTE:
            if sensorData.deviceType == DeviceType.NET:
                if "READ" in sensorData.id:
                    metricType = MetricType.NET_READ
                else:
                    metricType = MetricType.NET_WRITE
            else:
                if "READ" in sensorData.id:
                    metricType = MetricType.DISK_READ
                else:
                    metricType = MetricType.DISK_WRITE

            result = rawData.values[-1] - rawData.values[0]
            sensorData.metrics[metricType] = Metric(
                metricType,
                result.astype(rawData.v_md.dtype),
                rawData.v_md,
                AggregateType.SUM,
            )

        sensorData.processed = True
    return sensorData


def processDataNode(dataNode: DataNode, force_process=False) -> DataNode:
    """Recursively calculate metrics of the current nodes, and of child nodes if necessary.

    Args:
        dataNode (DataNode): Root of the DataNode structure
        force_process (bool, optional): If true, ignored processed flag in child DataNodes. Defaults to False.
    """
    aggregatedMetrics: Dict[MetricType, List[Metric]] = {}
    for _, subNode in dataNode.nodes.items():
        # Make sure sub nodes have their metrics ready
        if not subNode.processed or force_process:
            if subNode.type == NodeType.SENSOR:
                subNode = processSensorData(subNode)
            else:
                subNode = processDataNode(subNode, force_process=force_process)

        for metricType, metric in subNode.metrics.items():
            if isinstance(metric, Metric):
                if metricType in aggregatedMetrics:
                    aggregatedMetrics[metricType].append(metric)
                else:
                    aggregatedMetrics[metricType] = [metric]

    for metricType, metrics in aggregatedMetrics.items():
        aggType = metrics[0].agg
        metric_md = metrics[0].metric_md
        if dataNode.type == NodeType.MULTI_RUN:
            dataNode.metrics[metricType] = Stats.fromMetrics(metrics)
        else:
            if aggType == AggregateType.MEAN:
                aggregatedValue = np.array([metric.value for metric in metrics]).mean()
            elif aggType == AggregateType.MAX:
                aggregatedValue = np.array([metric.value for metric in metrics]).max()
            elif aggType == AggregateType.MIN:
                aggregatedValue = np.array([metric.value for metric in metrics]).min()
            else:
                aggregatedValue = np.array([metric.value for metric in metrics]).sum()

            dataNode.metrics[metricType] = Metric(
                metricType, aggregatedValue, metric_md, aggType
            )

    dataNode.processed = True
    return dataNode



"""Core perun functionality."""
import platform
import pprint as pp

# import sys
import time
import types
from datetime import datetime
from multiprocessing import Event, Process, Queue
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import numpy as np

from perun import COMM_WORLD, __version__, log
from perun.configuration import config
from perun.coordination import getLocalSensorRankConfiguration
from perun.data_model.data import DataNode, NodeType, RawData
from perun.data_model.measurement_type import Magnitude, MetricMetaData, Unit
from perun.data_model.sensor import DeviceType, Sensor
from perun.io.io import IOFormat, exportTo
from perun.processing import processDataNode, processSensorData
from perun.util import getRunId, getRunName


def monitor_application(
    app: Union[Path, Callable], app_args: tuple = tuple(), app_kwargs: dict = dict()
) -> Optional[Any]:
    """Execute coordination, monitoring, post-processing, and reporting steps, in that order.

    Raises:
        executionError: From cmd line execution, raises any error found on the monitored app.
        functionError:  From decorator execution, raises any error found on the monitored method.

    Returns:
        Optional[Any]: In decorator mode, return the output of the decorated method.
    """
    # outPath: Path = Path(config.get("output", "data_out"))

    # Get node devices
    from perun.backend import backends

    log.debug(f"Backends: {backends}")

    starttime = datetime.utcnow()
    app_results: Optional[Any] = None
    data_out = Path(config.get("output", "data_out"))
    format = IOFormat(config.get("output", "format"))
    includeRawData = config.getboolean("output", "raw")
    depthStr = config.get("output", "depth")
    if depthStr:
        depth = int(depthStr)
    else:
        depth = None

    if not config.getboolean("benchmarking", "bench_enable"):
        app_results, dataNode = _run_application(
            backends, app, app_args, app_kwargs, record=True
        )
        if dataNode:
            # Only on first rank, export data
            exportTo(data_out, dataNode, format, includeRawData, depth)
    else:
        # Start with warmup rounds
        log.info("Started warmup rounds")
        for i in range(config.getint("benchmarking", "bench_warmup_rounds")):
            log.debug(f"Warmup run: {i}")
            app_results, _ = _run_application(
                backends, app, app_args, app_kwargs, record=False
            )

        log.info("Started bench rounds")
        runNodes: List[DataNode] = []
        for i in range(config.getint("benchmarking", "bench_rounds")):
            log.debug(f"Bench run: {i}")
            app_results, runNode = _run_application(
                backends, app, app_args, app_kwargs, record=True, run_id=str(i)
            )
            if runNode:
                runNodes.append(runNode)

        if len(runNodes) > 0:
            benchNode = DataNode(
                id=getRunId(starttime),
                type=NodeType.MULTI_RUN,
                metadata={
                    "app_name": getRunName(app),
                    "startime": starttime.isoformat(),
                    "perun_version": __version__,
                },
                nodes={node.id: node for node in runNodes},
            )
            benchNode = processDataNode(benchNode)

            exportTo(data_out, benchNode, format, includeRawData, depth)

    for backend in backends:
        backend.close()

    return app_results


def _run_application(
    backends: List,
    app: Union[Path, Callable],
    app_args: tuple = tuple(),
    app_kwargs: dict = dict(),
    record: bool = True,
    run_id: Optional[str] = None,
) -> Tuple[Optional[Any], Optional[DataNode]]:
    app_result: Optional[Any] = None
    if record:
        # 1) Get sensor configuration
        mpiRanks, localBackends = getLocalSensorRankConfiguration(COMM_WORLD, backends)

        start_event = Event()
        stop_event = Event()

        queue: Optional[Queue] = None
        perunSP: Optional[Process] = None

        # 2) If assigned devices, create subprocess
        if len(localBackends.keys()) > 0:
            log.debug(
                f"Rank {COMM_WORLD.Get_rank()} - Local Backendens : {pp.pformat(localBackends)}"
            )
            queue = Queue()
            perunSP = Process(
                target=perunSubprocess,
                args=[
                    queue,
                    start_event,
                    stop_event,
                    localBackends,
                    config.getfloat("monitor", "sampling_rate"),
                ],
            )
            perunSP.start()

        # 3) Start application
        run_starttime = datetime.utcnow()

        if isinstance(app, Path):
            try:
                with open(str(app), "r") as scriptFile:
                    start_event.wait()
                    exec(
                        scriptFile.read(),
                        {"__name__": "__main__", "__file__": app.name},
                    )
                    stop_event.set()
            except Exception as e:
                log.error(f"Found error on monitored script: {str(app)}")
                stop_event.set()
                raise e

        elif isinstance(app, types.FunctionType):
            try:
                start_event.wait()
                app_result = app(*app_args, **app_kwargs)
            except Exception as e:
                stop_event.set()
                raise e
            stop_event.set()

        # 4) App finished, stop subrocess and get data
        if queue and perunSP:
            log.debug("Getting queue contents")
            nodeData = queue.get(block=True)
            log.debug("Got queue contents")
            log.debug("Waiting for subprocess to close")
            perunSP.join()
            perunSP.close()
            log.debug("Subprocess closed")
            queue.close()
        else:
            nodeData = None

        COMM_WORLD.barrier()
        log.debug("Everyone exited the subprocess")

        if nodeData:
            nodeData.metadata["mpi_ranks"] = mpiRanks

        # 5) Collect data from everyone on the first rank
        dataNodes: Optional[List[DataNode]] = COMM_WORLD.gather(nodeData, root=0)
        if dataNodes:
            # 6) On the first rank, create run node
            runNode = DataNode(
                id=run_id if run_id is not None else getRunId(run_starttime),
                type=NodeType.RUN,
                metadata={
                    "app_name": getRunName(app),
                    "startime": run_starttime.isoformat(),
                    "perun_version": __version__,
                },
                nodes={node.id: node for node in dataNodes if node},
            )
            runNode = processDataNode(runNode)

            return app_result, runNode
        return app_result, None

    else:
        if isinstance(app, Path):
            # filePath = app
            try:
                with open(str(app), "r") as scriptFile:
                    exec(
                        scriptFile.read(),
                        {"__name__": "__main__", "__file__": app.name},
                    )
            except Exception as e:
                log.error(f"Found error on monitored script: {str(app)}")
                raise e

        elif isinstance(app, types.FunctionType):
            # filePath = Path(sys.argv[0])
            try:
                app_result = app(*app_args, **app_kwargs)
            except Exception as e:
                raise e

        return app_result, None


def perunSubprocess(
    queue: Queue,
    start_event,
    stop_event,
    backendConfig: Dict[str, Set[str]],
    sampling_rate: float,
):
    """
    Parallel function that samples energy values from hardware libraries.

    Args:
        queue (Queue): multiprocessing Queue object where the results are sent after finish
        start_event (_type_): Marks that perun finished setting up and started sampling from devices
        stop_event (_type_): Signal the subprocess that the monitored processed has finished
        deviceIds (Set[str]): List of device ids to sample from
        sampling_rate (int): Sampling rate in s
    """
    from perun.backend import backends

    lSensors: List[Sensor] = []
    for backend in backends:
        backend.setup()
        if backend.name in backendConfig:
            lSensors += backend.getSensors(backendConfig[backend.name])

    timesteps = []
    t_mT = MetricMetaData(
        Unit.SECOND,
        Magnitude.ONE,
        np.dtype("float32"),
        np.float32(0),
        np.finfo("float32").max,
        np.float32(-1),
    )
    rawValues: List[List[np.number]] = []
    for _ in lSensors:
        rawValues.append([])

    log.debug(f"perunSP lSensors: {lSensors}")

    start_event.set()
    timesteps.append(time.time_ns())
    for idx, device in enumerate(lSensors):
        rawValues[idx].append(device.read())

    while not stop_event.wait(sampling_rate):
        timesteps.append(time.time_ns())
        for idx, device in enumerate(lSensors):
            rawValues[idx].append(device.read())

    timesteps.append(time.time_ns())
    for idx, device in enumerate(lSensors):
        rawValues[idx].append(device.read())

    log.debug("Subprocess: Stop event received.")
    for backend in backends:
        backend.close()

    log.debug("Subprocess: Closed backends")
    sensorNodes: Dict = {}

    t_s = np.array(timesteps)
    t_s -= t_s[0]
    t_s = t_s.astype("float32")
    t_s *= 1e-9

    for sensor, values in zip(lSensors, rawValues):
        if sensor.type not in sensorNodes:
            sensorNodes[sensor.type] = []

        dn = DataNode(
            id=sensor.id,
            type=NodeType.SENSOR,
            metadata=sensor.metadata,
            deviceType=sensor.type,
            raw_data=RawData(t_s, np.array(values), t_mT, sensor.dataType),
        )
        # Apply processing to sensor node
        dn = processSensorData(dn)
        sensorNodes[sensor.type].append(dn)

    log.debug("Subprocess: Preprocessed Sensor Data")
    deviceGroupNodes = []
    for deviceType, sensorNodes in sensorNodes.items():
        if deviceType != DeviceType.NODE:
            dn = DataNode(
                id=deviceType.value,
                type=NodeType.DEVICE_GROUP,
                metadata={},
                nodes={sensor.id: sensor for sensor in sensorNodes},
                deviceType=deviceType,
            )

            dn = processDataNode(dn)
            deviceGroupNodes.append(dn)
        else:
            deviceGroupNodes.extend(sensorNodes)

    log.debug("Subprocess: Preprocessed Device Data")

    hostNode = DataNode(
        id=platform.node(),
        type=NodeType.NODE,
        metadata={},
        nodes={node.id: node for node in deviceGroupNodes},
    )
    processDataNode(hostNode)

    # This should send a single processed node for the current computational node
    queue.put(hostNode, block=True)
    log.debug("Subprocess: Sent data")

    for backend in backends:
        backend.close()



"""Storage Module."""
import dataclasses
import enum
from typing import Any, Dict, List, Optional, Union

import numpy as np
from typing_extensions import Self

from perun import log
from perun.data_model.measurement_type import MetricMetaData
from perun.data_model.sensor import DeviceType


class NodeType(enum.Enum):
    """DataNode type enum."""

    MULTI_RUN = enum.auto()
    RUN = enum.auto()
    NODE = enum.auto()
    DEVICE_GROUP = enum.auto()
    SENSOR = enum.auto()


class MetricType(str, enum.Enum):
    """Metric Type enum."""

    RUNTIME = "runtime"
    POWER = "power"
    CPU_UTIL = "cpu_util"
    GPU_UTIL = "gpu_util"
    MEM_UTIL = "mem_util"
    NET_READ = "net_read"
    NET_WRITE = "net_write"
    DISK_READ = "disk_read"
    DISK_WRITE = "disk_write"
    ENERGY = "energy"


class AggregateType(str, enum.Enum):
    """Types of data aggregation."""

    SUM = "sum"
    MEAN = "mean"
    MAX = "max"
    MIN = "min"


@dataclasses.dataclass
class Metric:
    """Struct with resulting metrics and the metadata."""

    type: MetricType
    value: np.number
    metric_md: MetricMetaData
    agg: AggregateType

    @classmethod
    def fromDict(cls, metricDict: Dict) -> Self:
        """Create RawData object from a dictionary."""
        return cls(
            MetricType(metricDict["type"]),
            np.float32(metricDict["value"]),
            MetricMetaData.fromDict(metricDict["metric_md"]),
            AggregateType(metricDict["agg"]),
        )


@dataclasses.dataclass
class Stats:
    """Collects statistics based on multiple metrics of the same type."""

    type: MetricType
    metric_md: MetricMetaData
    mean: np.number
    std: np.number
    max: np.number
    min: np.number

    @classmethod
    def fromMetrics(cls, metrics: List[Metric]) -> Self:
        """Create a stats object based on the metric's values."""
        type = metrics[0].type
        metric_md = metrics[0].metric_md

        for m in metrics:
            if m.type != type:
                log.error("Metrics given to Stats class do not match")
                raise Exception("Metrics type don't match. Invalid Stats")

        values = np.array([metric.value for metric in metrics])
        mean = values.mean()
        std = values.std()
        max = values.max()
        min = values.min()
        return cls(type, metric_md, mean, std, max, min)

    @property
    def value(self):
        """
        Value property.

        For compatibility with Metric dataclass.
        """
        return self.mean

    @classmethod
    def fromDict(cls, statsDict: Dict) -> Self:
        """Stats constructor from a dictionory."""
        return cls(
            MetricType(statsDict["type"]),
            MetricMetaData.fromDict(statsDict["metric_md"]),
            statsDict["mean"],
            statsDict["std"],
            statsDict["max"],
            statsDict["min"],
        )


@dataclasses.dataclass
class RawData:
    """Contains timesteps and recorded values from sensors, including information on the values."""

    timesteps: np.ndarray
    values: np.ndarray
    t_md: MetricMetaData
    v_md: MetricMetaData

    @classmethod
    def fromDict(cls, rawDataDict: Dict) -> Self:
        """Create RawData object from a dictionary."""
        t_md = MetricMetaData.fromDict(rawDataDict["t_md"])
        v_md = MetricMetaData.fromDict(rawDataDict["v_md"])
        return cls(
            np.array(rawDataDict["timesteps"], dtype=t_md.dtype),
            np.array(rawDataDict["values"], dtype=t_md.dtype),
            t_md,
            v_md,
        )


class DataNode:
    """Recursive data structure that contains all the information of a monitored application."""

    def __init__(
        self,
        id: str,
        type: NodeType,
        metadata: Dict,
        nodes: Optional[Dict[str, Self]] = None,
        metrics: Optional[Dict[MetricType, Union[Metric, Stats]]] = None,
        deviceType: Optional[DeviceType] = None,
        raw_data: Optional[RawData] = None,
        processed: bool = False,
    ) -> None:
        """DataNode.

        Args:
            id (str): String identifier
            type (NodeType): Type of Node
            metadata (Dict): Metadata
            nodes (Dict[str, Self], optional): Child DataNodes. Defaults to {}.
            raw_data (Optional[RawData], optional): If sensor, contains raw sensor values. Defaults to None.
        """
        self.id = id
        self.type = type
        self.metadata: Dict[str, Any] = metadata
        self.nodes: Dict[str, Self] = nodes if nodes else {}
        self.metrics: Dict[MetricType, Union[Metric, Stats]] = (
            metrics if metrics else {}
        )
        self.deviceType: Optional[DeviceType] = deviceType
        self.raw_data: Optional[RawData] = raw_data
        self.processed = processed

    def toDict(
        self, depth: Optional[int] = None, include_raw_data: bool = False
    ) -> Dict:
        """Transform object to dictionary."""
        resultsDict = {
            "id": self.id,
            "type": self.type.value,
            "metadata": self.metadata,
            "metrics": {
                type.value: dataclasses.asdict(metric)
                for type, metric in self.metrics.items()
            },
            "deviceType": self.deviceType,
            "processed": self.processed,
        }
        if depth is None:
            resultsDict["nodes"] = (
                {
                    key: value.toDict(depth, include_raw_data)
                    for key, value in self.nodes.items()
                },
            )
        elif depth > 1:
            resultsDict["nodes"] = (
                {
                    key: value.toDict(depth - 1, include_raw_data)
                    for key, value in self.nodes.items()
                },
            )

        if include_raw_data and self.raw_data:
            resultsDict["raw_data"] = dataclasses.asdict(self.raw_data)

        return resultsDict

    @classmethod
    def fromDict(cls, resultsDict: Dict) -> Self:
        """Build object from dictionary."""
        type = NodeType(resultsDict["type"])
        newResults = cls(
            id=resultsDict["id"],
            type=type,
            metadata=resultsDict["metadata"],
            nodes={
                key: DataNode.fromDict(node)
                for key, node in resultsDict["nodes"].items()
            },
            processed=resultsDict["processed"],
        )
        if "deviceType" in resultsDict:
            newResults.deviceType = DeviceType(resultsDict["deviceType"])

        if "metrics" in resultsDict:
            if type == NodeType.MULTI_RUN:
                newResults.metrics = {
                    MetricType(type): Stats.fromDict(metricDict)
                    for type, metricDict in resultsDict["metrics"].items()
                }
            else:
                newResults.metrics = {
                    MetricType(type): Metric.fromDict(metricDict)
                    for type, metricDict in resultsDict["metrics"].items()
                }
        if "raw_data" in resultsDict:
            newResults.raw_data = RawData.fromDict(resultsDict["raw_data"])

        return newResults
